from dune.codegen.error import CodegenError
from dune.codegen.generation import quadrature_mixin
from dune.codegen.pdelab.quadrature import GenericQuadratureMixin
from dune.codegen.blockstructured.tools import name_point_in_macro, sub_element_inames, remove_sub_element_inames
import pymbolic.primitives as prim


@quadrature_mixin("blockstructured")
class BlockstructuredQuadratureMixin(GenericQuadratureMixin):
    @staticmethod
    def _subscript_without_sub_element_inames(s):
        return prim.Subscript(s.aggregate, remove_sub_element_inames(s.index_tuple))

    def quadrature_position_in_micro(self, index=None):
        return self._subscript_without_sub_element_inames(super().quadrature_position(index))

    def quadrature_position_in_macro(self, index=None):
        original = self.quadrature_position_in_micro(index)
        qp = prim.Variable(name_point_in_macro(original, self), )
        if index is not None:
            return prim.Subscript(qp, (index,))
        else:
            return qp

    def quadrature_position(self, index=None):
        raise CodegenError('Decide if the quadrature point is in the macro or micro element')

    def quadrature_inames(self):
        return super().quadrature_inames() + sub_element_inames()

    def quadrature_weight(self, o):
        return self._subscript_without_sub_element_inames(super().quadrature_weight(o))
